// 
// Copyright (C) 2005 Stefan Domnanovits
// mailto:stefan@domnanovits.at
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either version 2
// of the License, or (at your option) any later version.

// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details

#import "Apple Tree_Prefix.pch"
#include "ATDiskElement.h"

ATElementId ATDiskElement::_idSequence(0);

//-----------------------------------------------------------------------------
void ATDiskElement::ResetIdSequence()
//-----------------------------------------------------------------------------
{
	_idSequence = 0;
}

//-----------------------------------------------------------------------------
ATDiskElement::ATDiskElement() : _bytes(0), _name(nil), _pParent(NULL),
	_id(GetNextId())
//-----------------------------------------------------------------------------
{
}

//-----------------------------------------------------------------------------
ATDiskElement::ATDiskElement( const ATDiskElement & element ) :
	_bytes( element._bytes ), _name(nil), _subElements(element._subElements),
	_pParent(element._pParent), _id(element._id)
//-----------------------------------------------------------------------------
{
	Name( element.Name() );
}

//-----------------------------------------------------------------------------
ATDiskElement::ATDiskElement( float bytes, NSString * name ) :
_bytes(0), _name(nil), _pParent(NULL), _id(GetNextId())
//-----------------------------------------------------------------------------
{
	Bytes( bytes );
	Name( name );
}

//-----------------------------------------------------------------------------
void ATDiskElement::Name( NSString * name )
//-----------------------------------------------------------------------------
{
	if ( name )
		[name retain];

	[_name release];
	_name = name;
}

//-----------------------------------------------------------------------------
ATDiskElement::~ATDiskElement()
//-----------------------------------------------------------------------------
{
	[_name release];
}

//-----------------------------------------------------------------------------
void ATDiskElement::Bytes( float bytes )
//-----------------------------------------------------------------------------
{
	_bytes = bytes;
}

//-----------------------------------------------------------------------------
ATDiskElement & ATDiskElement::AddSubElement( NSString * name, float bytes )
//-----------------------------------------------------------------------------
{
	// The size of the current element is the sum of all subelements
	_bytes += bytes;
	
	// create new subelement
	ElementMap::iterator pos = _subElements.insert( 
						ElementMap::value_type(bytes, ATDiskElement(bytes, name)) );
	return pos->second;
}


//-----------------------------------------------------------------------------
ATDiskElement & ATDiskElement::AddSubElement( ATDiskElement & element )
//-----------------------------------------------------------------------------
{
	_bytes += element.Bytes();
	ElementMap::iterator pos = _subElements.insert( 
								ElementMap::value_type(element.Bytes(), element) );
	return pos->second;
}

//-----------------------------------------------------------------------------
void ATDiskElement::Clear()
//-----------------------------------------------------------------------------
{
	Name( nil );
	Bytes( 0 );
	_subElements.clear();
	_pathName = StringElemPtr(); 
}

//-----------------------------------------------------------------------------
std::string ATDiskElement::ToString()
//-----------------------------------------------------------------------------
{
	char buffer[20];
	std::string s;
	
	s += "count=";
	sprintf( buffer, "%d", _subElements.size() );
	s += buffer;
	s += "; size=";
	sprintf( buffer, "%d", static_cast<int>(_bytes) );
	s += buffer;
	s += "; ";
	s += [_name UTF8String];
	
	return s;
}

//-----------------------------------------------------------------------------
void ATDiskElement::WriteElementContent()
//-----------------------------------------------------------------------------
{
	for ( ElementMap::iterator i=begin(); i!=end(); ++i )
	{
		printf( " - %s\n", i->second.ToString().c_str() );
	}
}

//-----------------------------------------------------------------------------
void ATDiskElement::WriteParentData()
//-----------------------------------------------------------------------------
{
	for ( ElementMap::iterator i=begin(); i!=end(); ++i )
	{
		i->second.Parent( this );
		i->second.WriteParentData();
	}
}

//-----------------------------------------------------------------------------
void ATDiskElement::Parent( ATDiskElement * pParent )
//-----------------------------------------------------------------------------
{
	_pParent = pParent;
}

//-----------------------------------------------------------------------------
NSString * ATDiskElement::GetPathName()
//-----------------------------------------------------------------------------
{
	NSString * pathName = nil;
	if ( _pathName.get() )
	{
		pathName = static_cast<NSString*>( (*_pathName) );
	}
	else
	{
		if ( Parent() )
		{
			pathName = Parent()->GetPathName();
			pathName = [pathName stringByAppendingPathComponent:Name()];
		}
		else
		{
			pathName = Name();
		}
		_pathName = StringElemPtr( new ATStringContainer(pathName) );
	}
	
	if ( !IsDir() )
	{
		pathName = [pathName stringByAppendingString:@" - "];
		pathName = [pathName stringByAppendingString:MakeSizeString()];
	}
	
	return pathName;
}

//-----------------------------------------------------------------------------
NSString * ATDiskElement::MakeSizeString()
//-----------------------------------------------------------------------------
{
	NSString *ret = nil;
	if( _bytes == 0 ) ret = @"Zero bytes";
	else if( _bytes > 0 && _bytes < 1024 ) ret = [NSString stringWithFormat:@"%f bytes", _bytes];
	else if( _bytes >= 1024 && _bytes < pow(1024, 2) ) ret = [NSString stringWithFormat:@"%.1f KB", (_bytes / 1024 )];
	else if( _bytes >= pow(1024, 2) && _bytes < pow(1024, 3) ) ret = [NSString stringWithFormat:@"%.2f MB", (_bytes / pow(1024., 2))];
	else if( _bytes >= pow(1024, 3) && _bytes < pow(1024, 4) ) ret = [NSString stringWithFormat:@"%.3f GB", (_bytes / pow( 1024., 3))];
	else if( _bytes >= pow(1024, 4) ) ret = [NSString stringWithFormat:@"%.4f TB", (_bytes / pow(1024, 4))];
	return ret;
}

//-----------------------------------------------------------------------------
int ATDiskElement::GetNextId()
//-----------------------------------------------------------------------------
{
	return ++_idSequence;
}