// 
// Copyright (C) 2005 Stefan Domnanovits
// mailto:stefan@domnanovits.at
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either version 2
// of the License, or (at your option) any later version.

// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details

#import "Apple Tree_Prefix.pch"
#include "ATPainter.h"
#include <math.h>

const float STD_BACKGRND = 0.796875;

//-----------------------------------------------------------------------------
ATPainter::ATPainter( CGRect rect, CGContextRef context, Analyzer & analyzer, ATTreeView * view )
	: _analyzer(analyzer), _view(view)
//-----------------------------------------------------------------------------
{
	_mainRect = rect;
	_context  = context;
	_isLiveResize = [_view duringLiveResize]==YES ? true : false;
	
	if ( _isLiveResize )
	{
		_analyzer.ClearCoordinates();
	}
	
	NSPoint lastPoint = [_view lastPoint];
	_cursorPoint.x = lastPoint.x;
	_cursorPoint.y = lastPoint.y;
}

//-----------------------------------------------------------------------------
void ATPainter::DrawTree()
//-----------------------------------------------------------------------------
{
	if ( Data().HasData() )
	{
		CGContextSetShouldAntialias( Context(), false );
		DrawData();
		
		if ( _isLiveResize )
		{
			_analyzer.ClearCoordinates();
			[_view removeAllToolTips];
		}
		
		// Draw border frame
		CGContextSetRGBStrokeColor( Context(), 0, 0, 0, 1 );
		CGContextStrokeRect( Context(), GetMainMinRect() );
	}
	else
	{
		DrawBackGround( MainRect() );
		
		_analyzer.ClearCoordinates();
		[_view removeAllToolTips];
		CGContextSetShouldAntialias( Context(), true );
		DrawNoData();
	}
	

}

//-----------------------------------------------------------------------------
void ATPainter::DrawData()
//-----------------------------------------------------------------------------
{
	ATDiskElement & rootElement = Data().RootElement();
	
	if ( _analyzer.HasCoordinates() )
	{
		DrawFromCoordinates();
	}
	else
	{
		DrawBackGround( MainRect() );
		DrawElement( rootElement, MainRect() );
	}
}

//-----------------------------------------------------------------------------
void ATPainter::DrawElement( ATDiskElement & element, const CGRect & rect )
//-----------------------------------------------------------------------------
{
	//if ( HasDrawSize(rect) )
	{
		ElementMap::iterator begin = element.begin();
		ElementMap::iterator pos = element.begin();
		DrawDirection direction = GetDirection( rect );
	
		DrawElemVector drawElements;
		CGRect remainingRect;
		CGRect drawRect = rect;
		
		ATColor elementColor = ATColor::ElementDark();
		ATColor elementIncr = elementColor.GetGradientIncrement( ATColor::ElementLight(), element.size() );
	
		// Draw current element layer
		while ( pos != element.end() )
		{
			GetNextRowElements( begin, pos, element.end(), rect, direction );
			DrawRow( begin, pos, element.end(), drawRect, direction, remainingRect, drawElements, elementColor, elementIncr );
		
			direction = Inverse( direction );
		
			begin = pos;
			drawRect = remainingRect;
		}
		
		if ( !_isLiveResize )
		{
			// Draw subelemnts
			for ( DrawElemVector::iterator i=drawElements.begin(); i!=drawElements.end(); ++i )
			{
				DrawElement( i->Element()->second, i->Rect() );
			}
		}
	}
}

//-----------------------------------------------------------------------------
void ATPainter::DrawRow( /*[in]*/ ElementMap::iterator begin, 
								 /*[in]*/ ElementMap::iterator rowEnd,
								 /*[in]*/ ElementMap::iterator end,
								 /*[in]*/ CGRect & rect, 
								 /*[in]*/ DrawDirection dir,
								 /*[out]*/ CGRect & remainingRect,
								 /*[ref]*/ DrawElemVector & drawElements,
								 /*[ref]*/ ATColor & elementColor,
								 /*[in] */ const ATColor & incrColor )
//-----------------------------------------------------------------------------
{
	CGRect drawRect;
	float rel = GetRowChildRelation( begin, rowEnd, end );
	CalculateDrawRects( rect, rel, dir, drawRect, remainingRect );
	
	if ( dir == vertical )
	{
		DrawVerticalRow( begin, rowEnd, drawRect, drawElements, elementColor, incrColor );
	}
	else
	{
		DrawHorizontalRow( begin, rowEnd, drawRect, drawElements, elementColor, incrColor );
	}
}

//-----------------------------------------------------------------------------
void ATPainter::CalculateDrawRects( /*[in]*/ CGRect & rect, 
												/*[in]*/ float rel, 
												/*[in]*/ DrawDirection dir, 
												/*[in]*/ CGRect & drawRect, 
												/*[out]*/ CGRect & remainingRect )
//-----------------------------------------------------------------------------
{
	if ( dir == vertical )
	{
		drawRect.origin = rect.origin;
		drawRect.size.height = rect.size.height;
		drawRect.size.width = rect.size.width * rel;
		
		remainingRect.origin.y = rect.origin.y;
		remainingRect.origin.x = rect.origin.x + drawRect.size.width;
		remainingRect.size.height = rect.size.height;
		remainingRect.size.width = rect.size.width - drawRect.size.width;
	}
	else
	{
		drawRect.origin = rect.origin;
		drawRect.size.width = rect.size.width;
		drawRect.size.height = rect.size.height * rel;
		
		remainingRect.origin.x = rect.origin.x;
		remainingRect.origin.y = rect.origin.y + drawRect.size.height;
		remainingRect.size.width = rect.size.width;
		remainingRect.size.height = rect.size.height - drawRect.size.height;
	}
}

//-----------------------------------------------------------------------------
void ATPainter::DrawHorizontalRow( /*[in]*/ ElementMap::iterator begin,
											  /*[in]*/ ElementMap::iterator & end,
											  /*[in]*/ CGRect & rect,
											  /*[ref]*/ DrawElemVector & drawElements,
											  /*[ref]*/ ATColor & elementColor,
											  /*[in] */ const ATColor & incrColor )
//-----------------------------------------------------------------------------
{
	if ( begin!=end )
	{
		float sum = GetElementSum( begin, end );
		float offset = 0;
		CGRect drawRect = rect;
		do
		{
			drawRect.origin.x = rect.origin.x + offset;
			drawRect.size.width = rect.size.width * (begin->first / sum);
			
			_analyzer.Coordinates().AddCoordinateRect( drawRect, begin, elementColor );
			drawElements.push_back( ATDrawElement(drawRect, begin, elementColor) );
			
			if ( !begin->second.IsDir() || _isLiveResize )
				DrawRect( drawRect, elementColor );
			
			elementColor.IncremenColor( incrColor );
			
			offset += drawRect.size.width;
			++begin;
		} while ( begin != end );
	}
}

//-----------------------------------------------------------------------------
void ATPainter::DrawVerticalRow( /*[in]*/ ElementMap::iterator begin,
											/*[in]*/ ElementMap::iterator & end,
											/*[in]*/ CGRect & rect,
											/*[in]*/ DrawElemVector & drawElements,
											/*[ref]*/ ATColor & elementColor,
											/*[in] */ const ATColor & incrColor )
//-----------------------------------------------------------------------------
{
	if ( begin!=end )
	{
		float sum = GetElementSum( begin, end );
		float offset = 0;
		CGRect drawRect = rect;
		do
		{
			drawRect.origin.y = rect.origin.y + offset;
			drawRect.size.height = rect.size.height * (begin->first / sum);
			
			_analyzer.Coordinates().AddCoordinateRect( drawRect, begin, elementColor );
			drawElements.push_back( ATDrawElement(drawRect, begin, elementColor) );
			
			if ( !begin->second.IsDir() || _isLiveResize )
				DrawRect( drawRect, elementColor );
			
			elementColor.IncremenColor( incrColor );
		
			offset += drawRect.size.height;
			++begin;
		} while ( begin != end );
	}
}

//-----------------------------------------------------------------------------
void ATPainter::DrawRect( const CGRect & rect, const ATColor & color )
//-----------------------------------------------------------------------------
{
	if ( HasDrawSize(rect) )
	{	
		if ( _cursorPoint.x!=0 && _cursorPoint.y!=0 
			  && CGRectContainsPoint(rect, _cursorPoint) )
		{
			DrawFilledRectangle( rect, ATColor::ActiveBackGrnd() );
			DrawRectangle( rect, ATColor::ElementFrame() );
		}
		else
		{
			DrawFilledRectangle( rect, color );
			DrawRectangle( rect, ATColor::ElementFrame() );
		}
	}
}

//-----------------------------------------------------------------------------
float ATPainter::GetElementSum( ElementMap::iterator begin, ElementMap::iterator & end )
//-----------------------------------------------------------------------------
{
	float sum = 0;
	while ( begin!=end )
	{
		sum += begin->first;
		++begin;
	}
	return sum;
}

//-----------------------------------------------------------------------------
float ATPainter::GetWidth( CGRect & rect, DrawDirection dir )
//-----------------------------------------------------------------------------
{
	return (dir==vertical) ? rect.size.height : rect.size.width;
}

//-----------------------------------------------------------------------------
float ATPainter::GetRowChildRelation( ElementMap::iterator begin, ElementMap::iterator rowEnd, ElementMap::iterator & end )
//-----------------------------------------------------------------------------
{
	float sum = 0, rowSum = 0;
	return GetRowChildRelation( begin, rowEnd, end, sum, rowSum );
}

//-----------------------------------------------------------------------------
float ATPainter::GetRowChildRelation( /*[in]*/ ElementMap::iterator begin, 
												  /*[in]*/ ElementMap::iterator rowEnd, 
												  /*[in]*/ ElementMap::iterator & end,
												  /*[out]*/ float & sum,
												  /*[out]*/ float & rowSum )
//-----------------------------------------------------------------------------
{
	sum = 0;
	rowSum=0;
	
	for ( ; begin!=rowEnd; ++begin )
	{
		sum += begin->first;
		rowSum += begin->first;
	}
	
	for ( ;begin!=end; ++begin )
	{
		sum += begin->first;
	}
	
	return (sum!=0) ? rowSum/sum : 1;
}

//-----------------------------------------------------------------------------
void ATPainter::GetNextRowElements( /*[in]*/ ElementMap::iterator begin,
												/*[ref]*/ ElementMap::iterator & pos,
												/*[in]*/ ElementMap::iterator end,
												/*[in]*/ const CGRect & rect,
												/*[in]*/ DrawDirection dir )
//-----------------------------------------------------------------------------
{
	if ( pos!=end )
	{
		ElementMap::iterator next = ++pos;
		next++;
		
		float rel1 = GetRectRelation( begin, pos, end, rect, dir );
		float rel2 = GetRectRelation( begin, next, end, rect, dir );
		
		if ( DeltaRectRel(rel1) > DeltaRectRel(rel2) )
		{
			// continue plassing subelements in the current row
			GetNextRowElements( begin, pos, end, rect, dir );
		}
	}
}

//-----------------------------------------------------------------------------
float ATPainter::GetRectRelation( /*[in]*/ ElementMap::iterator begin, 
											 /*[in]*/ ElementMap::iterator rowEnd, 
											 /*[in]*/ ElementMap::iterator & end, 
											 /*[in]*/ const CGRect & rect,
											 /*[in]*/ DrawDirection dir )
//-----------------------------------------------------------------------------
{
	float rel = 1, sum, rowSum;
	float sizeRel = GetRowChildRelation( begin, rowEnd, end, sum, rowSum );
	float width, height;
	
	if ( dir == vertical )
	{
		height = rect.size.height;
		width = rect.size.width * sizeRel;
		
		rel = GetRectRel( --rowEnd, width, height, rowSum, dir );
	}
	else
	{
		height = rect.size.height * sizeRel;
		width = rect.size.width;
		
		rel = GetRectRel( --rowEnd, width, height, rowSum, dir );
	}
	
	return rel;
}

//-----------------------------------------------------------------------------
float ATPainter::DeltaRectRel( float rel )
//-----------------------------------------------------------------------------
{
	return fabsf( 1-rel );
}

//-----------------------------------------------------------------------------
void ATPainter::GetRowData( /*[in]*/ ElementMap::iterator start, 
									 /*[in]*/ ElementMap::iterator end, 
									 /*[out]*/ float & sum, 
									 /*[out]*/ float & minVal,
									 /*[out]*/ float & maxVal )
//-----------------------------------------------------------------------------
{
	sum = 0;
	minVal = INT_MAX;
	maxVal = 0;
	while ( start != end )
	{
		sum += start->first;
		maxVal = fmaxf( start->first, maxVal );
		minVal = fminf( start->first, minVal );
		++start;
	}
	
	if ( minVal == INT_MAX )
		minVal = 0;
}

//-----------------------------------------------------------------------------
void ATPainter::DrawXCenteredString( float y, NSString * text, NSString * font, NSColor * color, int size )
//-----------------------------------------------------------------------------
{
	NSMutableDictionary *attributes = [[NSMutableDictionary alloc] init];
	[attributes setObject:[NSFont fontWithName:font size:size]  
						forKey: NSFontAttributeName];
	[attributes setObject:color forKey:NSForegroundColorAttributeName];
		
	NSSize drawSize = [text sizeWithAttributes:attributes];
	
	NSPoint startPoint;
	startPoint.x = MainRect().origin.x + MainRect().size.width / 2 - drawSize.width / 2;
	startPoint.y = MainRect().origin.y + MainRect().size.height / 2 - drawSize.height / 2;
	
	[text drawAtPoint: startPoint
		withAttributes: attributes];
	
	[attributes release];
}

//-----------------------------------------------------------------------------
void ATPainter::DrawXCenteredStringAttribs( float y, NSString * text, NSMutableDictionary *attributes )
//-----------------------------------------------------------------------------
{
	NSSize drawSize = [text sizeWithAttributes:attributes];
	
	NSPoint startPoint;
	startPoint.x = MainRect().origin.x + MainRect().size.width / 2 - drawSize.width / 2;
	startPoint.y = y;
	
	[text drawAtPoint: startPoint
		withAttributes: attributes];
}

//-----------------------------------------------------------------------------
void ATPainter::DrawNoData()
//-----------------------------------------------------------------------------
{
	NSColor * color = [NSColor blackColor];
	NSMutableDictionary *attributes = [[NSMutableDictionary alloc] init];
	[attributes setObject:[NSFont fontWithName:@"Lucida Grande" size:14]  
						forKey:NSFontAttributeName];
	[attributes setObject:color forKey:NSForegroundColorAttributeName];
	
	NSString * subText = [NSString stringWithString:@"Please provide a valid directory and press Analyze."];
	NSSize drawSize = [subText sizeWithAttributes:attributes];
	
	float y = MainRect().origin.y + MainRect().size.height / 2 - drawSize.height / 2;
	
	DrawXCenteredString( y, @"No data available.", @"Lucida Grande", color, 20 );
	DrawXCenteredStringAttribs( y - drawSize.height, subText, attributes );
	
	[attributes release];
	
	// Draw border frame
	CGContextSetRGBStrokeColor( Context(), 0, 0, 0, 1 );
	CGContextStrokeRect( Context(), MainRect() );
}

//-----------------------------------------------------------------------------
DrawDirection ATPainter::GetDirection( const CGRect & rect )
//-----------------------------------------------------------------------------
{
	// We always draw along the smaler side
	return (rect.size.width > rect.size.height) ? vertical : horizontal;
}

//-----------------------------------------------------------------------------
float ATPainter::GetRectRel( /*[in]*/ ElementMap::iterator pElement, 
									  /*[in]*/ float width,
									  /*[in]*/ float height,
									  /*[in]*/ float sum,
									  /*[in]*/ DrawDirection dir )
//-----------------------------------------------------------------------------
{
	float rectRel = 0;
	float rel = pElement->first / sum;
	
	if ( dir == vertical )
	{
		float rectHeight = height * rel;
		rectRel = width/rectHeight;
	}
	else
	{
		float rectWidth = width * rel;
		rectRel = rectWidth/height;
	}
	
	return rectRel;
}

//-----------------------------------------------------------------------------
DrawDirection ATPainter::Inverse( DrawDirection dir )
//-----------------------------------------------------------------------------
{
	return dir==vertical ? horizontal : vertical;
}

//-----------------------------------------------------------------------------
inline bool ATPainter::HasDrawSize( const CGRect & rect )
//-----------------------------------------------------------------------------
{
	return ( rect.size.width > 0.5 || rect.size.height > 0.5 );
}

//-----------------------------------------------------------------------------
NSRect ATPainter::NSRectMake( const CGRect & rect )
//-----------------------------------------------------------------------------
{
	NSRect nsRect;
	
	nsRect.origin.x = rect.origin.x;
	nsRect.origin.y = rect.origin.y;
	nsRect.size.width = rect.size.width;
	nsRect.size.height = rect.size.height;
	
	return nsRect;
}

//-----------------------------------------------------------------------------
void ATPainter::DrawFromCoordinates()
//-----------------------------------------------------------------------------
{
	const NSRect *rects;
	int count = 0;
	[_view getRectsBeingDrawn:&rects count:&count];
	
	NSPoint drawPoint = [_view drawPoint];
	NSPoint lastPoint = [_view lastPoint];
	
	CoordContainer::iterator element;
	
	bool valid = _analyzer.GetPointElement( drawPoint, element );
	if ( valid )
	{
		DrawRect( element->second.Rect(), element->second.Color() );
		DrawParentRects( element->second.Element()->second, false );
	}
	
	valid = _analyzer.GetPointElement( lastPoint, element );
	if ( valid )
	{
		DrawRect( element->second.Rect(), element->second.Color() );
		DrawParentRects( element->second.Element()->second, true );
	}
}

//-----------------------------------------------------------------------------
void ATPainter::DrawParentRects( ATDiskElement & element, bool active )
//-----------------------------------------------------------------------------
{
	if ( element.Parent() )
	{
		CoordContainer::iterator drawElement;
		if ( _analyzer.GetDrawElement(element.Id(), drawElement) )
		{
			if ( active )
			{
				DrawRectangle( drawElement->second.Rect(), ATColor::ActiveElement() );
			}
			else
			{
				DrawRectangle( drawElement->second.Rect(), ATColor::ElementFrame() );
			}
			
			DrawParentRects( *(element.Parent()), active );
		}
	}
}

//-----------------------------------------------------------------------------
void ATPainter::DrawBackGround( const CGRect & rect )
//-----------------------------------------------------------------------------
{
	DrawFilledRectangle( rect, ATColor::BackGrnd() );
}

//-----------------------------------------------------------------------------
inline CGRect ATPainter::GetMainMinRect()
//-----------------------------------------------------------------------------
{
	return CGRectMake( MainRect().origin.x, MainRect().origin.y,
							 MainRect().size.width-1, MainRect().size.height-1 );
}

//-----------------------------------------------------------------------------
void ATPainter::DrawRectangle( const CGRect & rect, const ATColor & color )
//-----------------------------------------------------------------------------
{
	CGContextSetRGBStrokeColor( Context(), color.Red(), color.Green(), 
										 color.Blue(), color.Alpha() );
	CGContextStrokeRect( Context(), rect );
}

//-----------------------------------------------------------------------------
void ATPainter::DrawFilledRectangle( const CGRect & rect, const ATColor & color )
//-----------------------------------------------------------------------------
{
	CGContextSetRGBFillColor( Context(), color.Red(), color.Green(), 
									  color.Blue(), color.Alpha() );
	CGContextFillRect( Context(), rect );
}