// 
// Copyright (C) 2005 Stefan Domnanovits
// mailto:stefan@domnanovits.at
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either version 2
// of the License, or (at your option) any later version.

// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details

#import "Apple Tree_Prefix.pch"
#include "Analyzer.h"

//-----------------------------------------------------------------------------
Analyzer::Analyzer()
//-----------------------------------------------------------------------------
{
	_hasData = false;
}

//-----------------------------------------------------------------------------
void Analyzer::Analyze( NSString * path )
//-----------------------------------------------------------------------------
{
	NSLog( @"ATAnalyzer::analyze %@", path );
	Clear();

	NSFileManager * fileManager = [NSFileManager defaultManager];
	
	_rootElement.Name( path );
	int elementCount = AddSubElements( fileManager, _rootElement, _rootElement.Name() );
	NSLog( @"Number of elements: %d", elementCount );
	
	if ( DoCancel() )
	{
		Clear();
	}
	else
	{
		_rootElement.WriteParentData();
		_hasData = true;
	}
	_stringDict.Clear();
}

//-----------------------------------------------------------------------------
int Analyzer::AddSubElements( NSFileManager * fileManager, ATDiskElement & rootElement, NSString * path )
//-----------------------------------------------------------------------------
{
	NSAutoreleasePool *pool = [[NSAutoreleasePool alloc] init];
	int elementCount = 0;
	NSString * subElement;
	NSDirectoryEnumerator * dirEnum = [fileManager enumeratorAtPath:path];
	
	while ( (subElement = [dirEnum nextObject]) && !DoCancel() )
	{
		NSDictionary * attribs = [dirEnum fileAttributes];
		if ( !IsLink(attribs) )
		{
			++elementCount;
			
			if ( IsDirectory(attribs) )
			{
				[dirEnum skipDescendents];
				
				ATDiskElement newElement;
				newElement.Name( _stringDict.GetString(subElement) );
				elementCount += AddSubElements( fileManager, newElement, [path stringByAppendingPathComponent:subElement] );
				
				if ( newElement.Bytes() > 0.1 )
				{
					rootElement.AddSubElement( newElement );
				}
			}
			else
			{
				float bytes = GetSize( attribs );
				if ( bytes > 0.1 )
				{ 
					ATDiskElement sumElem = rootElement.AddSubElement( _stringDict.GetString(subElement), bytes );
				}
			}
		}
	}
	
	[pool release];
	return elementCount;
}

//-----------------------------------------------------------------------------
void Analyzer::Clear()
//-----------------------------------------------------------------------------
{
	_doCancel = false;
	_hasData = false;
	ClearCoordinates();
	_rootElement.Clear();
	_stringDict.Clear();
	ATDiskElement::ResetIdSequence();
}

//-----------------------------------------------------------------------------
float Analyzer::GetSize( NSDictionary * attribs )
//-----------------------------------------------------------------------------
{
	return [attribs fileSize];
}

//-----------------------------------------------------------------------------
bool Analyzer::IsLink( NSDictionary * attribs )
//-----------------------------------------------------------------------------
{
	bool isLink = false;
	
	NSString * fileType = [attribs fileType];
	if ( [fileType isEqualToString:NSFileTypeSymbolicLink] )
	{
		isLink = true;
	}
	
	return isLink;
}

//-----------------------------------------------------------------------------
bool Analyzer::IsDirectory( NSDictionary * attribs )
//-----------------------------------------------------------------------------
{
	bool isDir = false;
	
	NSString * fileType = [attribs fileType];
	if ( [fileType isEqualToString:NSFileTypeDirectory] )
	{
		isDir = true;
	}
	
	return isDir;
}


//-----------------------------------------------------------------------------
void Analyzer::CancelOperation()
//-----------------------------------------------------------------------------
{
	_doCancel = true;
}

//-----------------------------------------------------------------------------
ATDiskElement & Analyzer::RootElement()
//-----------------------------------------------------------------------------
{
	return _rootElement;
}

//-----------------------------------------------------------------------------
void Analyzer::ClearCoordinates()
//-----------------------------------------------------------------------------
{
	_coordinates.clear();
}

//-----------------------------------------------------------------------------
bool Analyzer::HasCoordinates()
//-----------------------------------------------------------------------------
{
	return !_coordinates.empty();
}

//-----------------------------------------------------------------------------
ATCoordinateMap & Analyzer::Coordinates()
//-----------------------------------------------------------------------------
{
	return _coordinates;
}

//-----------------------------------------------------------------------------
CGRect Analyzer::GetPointRect( const NSPoint & p )
//-----------------------------------------------------------------------------
{
	CGPoint p2 = CGPointMake( p.x, p.y );
	return GetPointRect( p2 );
}

//-----------------------------------------------------------------------------
CGRect Analyzer::GetPointRect( const CGPoint & p )
//-----------------------------------------------------------------------------
{
	CGRect rect = CGRectMake( 0, 0, 0, 0 );
	
	ATCoordinateMap & coords = Coordinates();
	CoordContainer::iterator pos = coords.FindElement( p );
	if ( pos != coords.end() )
	{
		const CGRect & r = pos->second.Rect();
		rect = r;
	}
	
	return rect;
}

//-----------------------------------------------------------------------------
bool Analyzer::GetPointElement( const NSPoint & p, CoordContainer::iterator & element )
//-----------------------------------------------------------------------------
{
	bool valid = false;
	
	ATCoordinateMap & coords = Coordinates();
	CoordContainer::iterator pos = coords.FindElement( p );
	if ( pos != coords.end() )
	{
		element = pos;
		valid = true;
	}
	
	return valid;
}

//-----------------------------------------------------------------------------
bool Analyzer::GetDrawElement( const ATElementId & elemenId, CoordContainer::iterator & drawElem )
//-----------------------------------------------------------------------------
{
	bool valid = false;
	
	ATCoordinateMap & coords = Coordinates();
	CoordContainer::iterator pos = coords.FindElement( elemenId );
	if ( pos != coords.end() )
	{
		drawElem = pos;
		valid = true;
	}
	
	return valid;
}
